// Copyright 2019 RnD Center "ELVEES", JSC

#include "utils.h"

#include <tinyxml2.h>

#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <cstring>
#include <sstream>
#include <utility>
#include <algorithm>

#include "cvatparse.h"

void parse_points(const std::string &points,
                  std::vector<cvatparse::point> &vpoints) {
    std::istringstream iss(points);
    cvatparse::point p;
    char comma, semicolon;
    while (!iss.eof()) {
        iss >> p.x >> comma >> p.y >> semicolon;
        vpoints.push_back(std::move(p));
    }
}

std::map<std::string, std::string> parse_attr(tinyxml2::XMLElement *xml_attr) {
    std::map<std::string, std::string> res_attr;
    tinyxml2::XMLElement *pattr = xml_attr->FirstChildElement("attribute");
    while (pattr != nullptr) {
        res_attr.emplace(std::make_pair(std::string(pattr->Attribute("name")),
                                        std::string(pattr->GetText())));
        pattr = pattr->NextSiblingElement("attribute");
    }
    return res_attr;
}

void parse_image_elem(tinyxml2::XMLElement *pelement,
                      std::vector<cvatparse::points> &meta) {
    std::vector<cvatparse::point> vpoints;
    parse_points(pelement->Attribute("points"), vpoints);
    meta.emplace_back(pelement->Attribute("label"),
                      vpoints,
                      parse_attr(pelement),
                      pelement->UnsignedAttribute("occluded"),
                      pelement->UnsignedAttribute("z_order"));
}

void parse_track_elem(tinyxml2::XMLElement *pelement,
                      std::vector<cvatparse::points> &meta) {
    std::vector<cvatparse::point> vpoints;
    parse_points(pelement->Attribute("points"), vpoints);
    meta.emplace_back(pelement->UnsignedAttribute("frame"),
                      vpoints,
                      parse_attr(pelement),
                      pelement->UnsignedAttribute("occluded"),
                      pelement->UnsignedAttribute("outside"),
                      pelement->UnsignedAttribute("keyframe"));
}

void parse_image_box(tinyxml2::XMLElement *pbox,
                     std::vector<cvatparse::box> &meta) {
    meta.emplace_back(pbox->Attribute("label"),
                      pbox->FloatAttribute("xtl"),
                      pbox->FloatAttribute("ytl"),
                      pbox->FloatAttribute("xbr"),
                      pbox->FloatAttribute("ybr"),
                      parse_attr(pbox),
                      pbox->UnsignedAttribute("occluded"),
                      pbox->UnsignedAttribute("z_order"));
}

void parse_track_box(tinyxml2::XMLElement *pbox,
                     std::vector<cvatparse::box> &meta) {
    meta.emplace_back(pbox->UnsignedAttribute("frame"),
                      pbox->FloatAttribute("xtl"),
                      pbox->FloatAttribute("ytl"),
                      pbox->FloatAttribute("xbr"),
                      pbox->FloatAttribute("ybr"),
                      parse_attr(pbox),
                      pbox->UnsignedAttribute("occluded"),
                      pbox->UnsignedAttribute("outside"),
                      pbox->UnsignedAttribute("keyframe"));
}

void parse(tinyxml2::XMLElement *xml_meta, cvatparse::image &meta) {
    tinyxml2::XMLElement *pelement = xml_meta->FirstChildElement();
    while (pelement != nullptr) {
        if (!std::strcmp(pelement->Name(), "box")) {
            parse_image_box(pelement, meta.box);
        } else if (!std::strcmp(pelement->Name(), "points")) {
            parse_image_elem(pelement, meta.points);
        } else if (!std::strcmp(pelement->Name(), "polygon")) {
            parse_image_elem(pelement, meta.polygon);
        } else if (!std::strcmp(pelement->Name(), "polyline")) {
            parse_image_elem(pelement, meta.polyline);
        }
        pelement = pelement->NextSiblingElement();
    }
}

void parse(tinyxml2::XMLElement *xml_meta, cvatparse::track &meta) {
    tinyxml2::XMLElement *pelement = xml_meta->FirstChildElement();
    while (pelement != nullptr) {
        if (!std::strcmp(pelement->Name(), "box")) {
            parse_track_box(pelement, meta.box);
        } else if (!std::strcmp(pelement->Name(), "points")) {
            parse_track_elem(pelement, meta.points);
        } else if (!std::strcmp(pelement->Name(), "polygon")) {
            parse_track_elem(pelement, meta.polygon);
        } else if (!std::strcmp(pelement->Name(), "polyline")) {
            parse_track_elem(pelement, meta.polyline);
        }
        pelement = pelement->NextSiblingElement();
    }
}

bool load_xml(tinyxml2::XMLDocument &xml_doc, const std::string &path) {
    tinyxml2::XMLError file_loaded = xml_doc.LoadFile(path.c_str());
    if (file_loaded != tinyxml2::XML_SUCCESS) {
        std::cout << "Error while loading file " << path << std::endl;
        return false;
    }
    return true;
}

std::string points_string(const std::vector<cvatparse::point> &vpoints) {
    if (vpoints.size() == 0) {
        return "";
    }
    std::ostringstream ss;
    for (auto const &pt : vpoints) {
        ss << pt.x << ',' << pt.y << ';';
    }
    std::string spoints(ss.str());
    spoints.pop_back();
    return spoints;
}

void serialize_attr(tinyxml2::XMLDocument &xml_doc,
                    tinyxml2::XMLElement *ppoints,
                    const std::map<std::string, std::string> &attrs) {
    for (auto const &attr : attrs) {
        tinyxml2::XMLElement *pattr = xml_doc.NewElement("attribute");
        pattr->SetAttribute("name", attr.first.c_str());
        pattr->SetText(attr.second.c_str());
        ppoints->InsertEndChild(pattr);
    }
}

void serialize_image_elem(tinyxml2::XMLDocument &xml_doc,
                          tinyxml2::XMLElement *xml_meta,
                          const std::vector<cvatparse::points> &meta,
                          const std::string &elem_name) {
    for (auto const &elem : meta) {
        tinyxml2::XMLElement *ppoints = xml_doc.NewElement(elem_name.c_str());
        ppoints->SetAttribute("label", elem.label.c_str());
        ppoints->SetAttribute("z_order", elem.z_order);
        ppoints->SetAttribute("points", points_string(elem.point).c_str());
        ppoints->SetAttribute("occluded", elem.occluded);
        serialize_attr(xml_doc, ppoints, elem.attr);
        xml_meta->InsertEndChild(ppoints);
    }
}

void serialize_track_elem(tinyxml2::XMLDocument &xml_doc,
                          tinyxml2::XMLElement *xml_meta,
                          const std::vector<cvatparse::points> &meta,
                          const std::string &elem_name) {
    for (auto const &elem : meta) {
        tinyxml2::XMLElement *ppoints = xml_doc.NewElement(elem_name.c_str());
        ppoints->SetAttribute("frame", elem.frame);
        ppoints->SetAttribute("keyframe", elem.keyframe);
        ppoints->SetAttribute("outside", elem.outside);
        ppoints->SetAttribute("points", points_string(elem.point).c_str());
        ppoints->SetAttribute("occluded", elem.occluded);
        serialize_attr(xml_doc, ppoints, elem.attr);
        xml_meta->InsertEndChild(ppoints);
    }
}

void serialize_image_box(tinyxml2::XMLDocument &xml_doc,
                         tinyxml2::XMLElement *xml_meta,
                         const std::vector<cvatparse::box> &meta) {
    for (auto const &box : meta) {
        tinyxml2::XMLElement *pbox = xml_doc.NewElement("box");
        pbox->SetAttribute("label", box.label.c_str());
        pbox->SetAttribute("z_order", box.z_order);
        pbox->SetAttribute("xtl", box.tl.x);
        pbox->SetAttribute("ytl", box.tl.y);
        pbox->SetAttribute("xbr", box.br.x);
        pbox->SetAttribute("ybr", box.br.y);
        pbox->SetAttribute("occluded", box.occluded);
        serialize_attr(xml_doc, pbox, box.attr);
        xml_meta->InsertEndChild(pbox);
    }
}

void serialize_track_box(tinyxml2::XMLDocument &xml_doc,
                         tinyxml2::XMLElement *xml_meta,
                         const std::vector<cvatparse::box> &meta) {
    for (auto const &box : meta) {
        tinyxml2::XMLElement *pbox = xml_doc.NewElement("box");
        pbox->SetAttribute("frame", box.frame);
        pbox->SetAttribute("keyframe", box.keyframe);
        pbox->SetAttribute("outside", box.outside);
        pbox->SetAttribute("xtl", box.tl.x);
        pbox->SetAttribute("ytl", box.tl.y);
        pbox->SetAttribute("xbr", box.br.x);
        pbox->SetAttribute("ybr", box.br.y);
        pbox->SetAttribute("occluded", box.occluded);
        serialize_attr(xml_doc, pbox, box.attr);
        xml_meta->InsertEndChild(pbox);
    }
}

void load_elems(tinyxml2::XMLElement *pimage, std::vector<cvatparse::image> &images) {
    cvatparse::image image_meta;
    while (pimage != nullptr) {
        image_meta.id = pimage->UnsignedAttribute("id");
        image_meta.height = pimage->UnsignedAttribute("height");
        image_meta.width = pimage->UnsignedAttribute("width");
        image_meta.name = pimage->Attribute("name");
        parse(pimage, image_meta);
        images.push_back(std::move(image_meta));
        pimage = pimage->NextSiblingElement("image");
    }
}

void load_elems(tinyxml2::XMLElement *ptrack, std::vector<cvatparse::track> &tracks) {
    cvatparse::track track_meta;
    while (ptrack != nullptr) {
        track_meta.id = ptrack->UnsignedAttribute("id");
        track_meta.label = ptrack->Attribute("label");
        parse(ptrack, track_meta);
        tracks.push_back(std::move(track_meta));
        ptrack = ptrack->NextSiblingElement("track");
    }
}

void dump_elems(tinyxml2::XMLDocument &xml_doc,
                tinyxml2::XMLNode *pannotations,
                std::vector<cvatparse::image> &images) {
    std::sort(images.begin(), images.end(),
              [] (cvatparse::image const& a, cvatparse::image const& b)
                  { return a.name < b.name; });
    for (unsigned int i = 0; i < images.size(); i++) {
        images[i].id = i;
    }
    for (auto const &im : images) {
        tinyxml2::XMLElement *pimage = xml_doc.NewElement("image");
        pimage->SetAttribute("id", im.id);
        pimage->SetAttribute("name", im.name.c_str());
        pimage->SetAttribute("height", im.height);
        pimage->SetAttribute("width", im.width);
        serialize_image_box(xml_doc, pimage, im.box);
        serialize_image_elem(xml_doc, pimage, im.points, "points");
        serialize_image_elem(xml_doc, pimage, im.polygon, "polygon");
        serialize_image_elem(xml_doc, pimage, im.polyline, "polyline");
        pannotations->InsertEndChild(pimage);
    }
}

void dump_elems(tinyxml2::XMLDocument &xml_doc,
                tinyxml2::XMLNode *pannotations,
                const std::vector<cvatparse::track> &tracks) {
    for (auto const &track : tracks) {
        tinyxml2::XMLElement *ptrack = xml_doc.NewElement("track");
        ptrack->SetAttribute("id", track.id);
        ptrack->SetAttribute("label", track.label.c_str());
        serialize_track_box(xml_doc, ptrack, track.box);
        serialize_track_elem(xml_doc, ptrack, track.points, "points");
        serialize_track_elem(xml_doc, ptrack, track.polygon, "polygon");
        serialize_track_elem(xml_doc, ptrack, track.polyline, "polyline");
        pannotations->InsertEndChild(ptrack);
    }
}
